/*

*/
package pipeline;
import scene.*;
import framebuffer.*;

import java.awt.Color;

/**
<pre>
   Rasterize each line segment into pixels in the framebuffer.

   This rasterization algorithm is based on
     "Fundamentals of Computer Graphics", 3rd Edition,
      by Peter Shirley, pages 163-165.
</pre>
*/
public class Rasterize
{
   public static boolean debug = false;

   /**
      Rasterize each line segment into pixels in the framebuffer.

      @param ls LineSegment to rasterize into the framebuffer
      @param fb FrameBuffer to hold rasterized pixels
   */
   public static void rasterize(LineSegment ls, FrameBuffer fb)
   {
      Color c = Color.white;

      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      // Round the line segment's two endpoints to the nearest
      // logical pixel. This makes the algorithm a lot simpler,
      // but it can cause a slight, but noticable, shift of the
      // line segment.
      double x0 = Math.round( ls.v[0].x );
      double y0 = Math.round( ls.v[0].y );
      double x1 = Math.round( ls.v[1].x );
      double y1 = Math.round( ls.v[1].y );

      if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
      {
         if (x1 < x0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute this line segment's slope.
         double m = (y1 - y0) / (x1 - x0);

         // Rasterize this line segment.
         // In the following loop, as x moves across the logical
         // horizontal pixels, we will compute a y value for each x.
         double y = y0;
         for (int x = (int)x0; x <= (int)x1; x++, y += m)
         {
            // The value of y will almost always be between
            // two vertical pixel coordinates. By rounding off
            // the value of y, we are choosing the nearest logical
            // vertical pixel coordinate.
            fb.setPixelVP(x - 1, h - (int)Math.round(y), c);

            // log interesting information to standard output
            if (debug) logPixel(w, h, x, y);

            // Advance (x,y) to the next pixel. Since delta_x = 1, we need delta_y = m.
         }
      }
      else  // abs(slope) > 1
      {
         if (y1 < y0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute this line segment's slope.
         double m = (x1 - x0) / (y1 - y0);

         // Rasterize this line segment.
         // In the following loop, as y moves across the logical
         // vertical pixels, we will compute a x value for each y.
         double x = x0;
         for (int y = (int)y0; y <= (int)y1; x += m, y++)
         {
            // The value of x will almost always be between
            // two horizontal pixel coordinates. By rounding off
            // the value of x, we are choosing the nearest logical
            // horizontal pixel coordinate.
            fb.setPixelVP((int)Math.round(x) - 1, h - y, c);

            // log interesting information to standard output
            if (debug) logPixel(w, h, x, y);

            // Advance (x,y) to the next pixel. Since delta_y = 1, we need delta_x = m.
         }
      }
   }


   private static void logPixel(int w, int h, int x, double y)
   {
      System.out.printf("[w=%d, h=%d]  x=%d, y=%f\n", w, h, x, y);
   }

   private static void logPixel(int w, int h, double x, int y)
   {
      System.out.printf("[w=%d, h=%d]  x=%f, y=%d\n", w, h, x, y);
   }
}
